#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

import asyncio
import json
from typing import Any

from pydantic import ConfigDict, Field

from xposer.api.base.base_kafka_service import BaseKafkaService
from xposer.api.base.base_kafka_service_config_model import BaseKafkaServiceConfigModel
from xposer.api.base.xpcontroller_base_class import XPControllerBaseClass
from xposer.core.configure import Configurator
from xposer.core.context import Context
from xposer.sample_app.rpc_kafka.sample_app_kafka_service import SampleAppKafkaService


class SampleAppKafkaXPControllerConfigModel(BaseKafkaServiceConfigModel):
    foo: str = Field(
        default='foo',
        description="Some xpcontroller specific config"
        )
    bar: str = Field('bar')
    model_config = ConfigDict(extra='allow')


class SampleAppKafkaXPController(XPControllerBaseClass):
    config_prefix: str = "xpcontroller_"

    def __init__(self, ctx: Context):
        super().__init__(ctx)
        self.config_prefix: str = "xpcontroller_"
        self.kafka_router: BaseKafkaService | Any = None
        self.config: SampleAppKafkaXPControllerConfigModel = self.config  # Type hint

    def mergeConfigurationFromPrefix(self) -> SampleAppKafkaXPControllerConfigModel:
        return Configurator.mergeAttributesWithPrefix(
            SampleAppKafkaXPControllerConfigModel,
            self.ctx.config,
            self.config_prefix,
            validate=True,
            strict=True
            )

    async def RPCHandler(self, data: Any):
        self.ctx.logger.info(
            f"Sample call with correlation id:{data.get('correlation_id', 'None')} receives sample raw data:\n"
            f"{json.dumps(data, indent=4)}"
            )
        return json.dumps({"result": "whoa", "originalfoo": data.get('foo', 'None')})

    async def start_kafka_service(self, callback):
        try:
            await self.kafka_service.start_service(
                server_string=self.config.router_kafka_server_string,
                group_id=self.config.router_kafka_group_id,
                inbound_topic=self.config.router_kafka_inbound_topic,
                outbound_topic=self.config.router_kafka_outbound_topic,
                exception_topic=self.config.router_kafka_exception_topic,
                handler_func=self.RPCHandler,
                produce_on_result=True
                )
            callback(None)
        except Exception as e:
            # Log the exception for debugging
            self.ctx.logger.exception(f"Error starting kafka service: {e}")
            # Notify the future object about the failure
            raise

    def handle_timeout_exception(self, task):
        try:
            task.result()
        except asyncio.TimeoutError:
            raise ValueError("The service did not start within 30 seconds!")

    async def tearDownXPController(self):
        ...

    async def startXPController(self):
        self.kafka_service = SampleAppKafkaService(self.ctx)
        future = asyncio.Future()
        kafka_service_task = asyncio.create_task(self.start_kafka_service(callback=future.set_result))
        kafka_service_task.set_name("SampleAppKafkaXPController:KafkaServiceTask")
        timeout_task = asyncio.create_task(asyncio.wait_for(future, timeout=3))
        timeout_task.set_name("SampleAppKafkaXPController:KafkaServiceTimeoutTask")
        timeout_task.add_done_callback(self.handle_timeout_exception)
        await future
        self.ctx.logger.debug("Started")



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

import asyncio
import json
from typing import Any

from pydantic import Field
from pydantic_settings import BaseSettings

from xposer.core.boot import Boot
from xposer.core.configure import Configurator
from xposer.core.context import Context


class SampleAppKafkaConfigModel(BaseSettings):
    logic_param: str = Field(default="logic_param_example_default_value")
    logic_param_to_override: str = Field(default="not_overridden")


class SampleAppKafka:
    ctx: Context
    config: SampleAppKafkaConfigModel
    config_prefix: str = "xpapp_"

    def __init__(self, ctx: Context):
        self.ctx = ctx
        self.config = Configurator.mergeAttributesWithPrefix(
            SampleAppKafkaConfigModel,
            ctx.config,
            self.config_prefix,
            validate=True,
            strict=True
            )
        self.ctx.logger.info(f"Initialized application")

    async def RPCHandler(self, data: Any):
        self.ctx.logger.info(
            f"Sample call with correlation id:{data.get('correlation_id', 'None')} receives sample raw data:\n"
            f"{json.dumps(data, indent=4)}"
            )
        return json.dumps({"result": "whoa", "originalfoo": data.get('foo', 'None')})


async def main():
    await Boot().boot()


if __name__ == "__main__":
    try:
        asyncio.run(main())
    except RuntimeError as e:
        if str(e) != 'Event loop stopped before Future completed.':
            raise



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

# In your main module
import asyncio

from xposer.core.boot import Boot


async def main():
    try:
        await Boot().boot()
    except SystemExit:
        pass  # Handle graceful shutdown here if needed


if __name__ == "__main__":
    try:
        asyncio.run(main())
    except SystemExit:
        pass  # Handle graceful shutdown here if needed



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

import json
import logging
import os
import sys

from confluent_kafka import Producer
from icecream import IceCreamDebugger

from xposer.core.configuration_model import ConfigModel

currentframe = lambda: sys._getframe(3)
_logging_srcfile = os.path.normcase(logging.addLevelName.__code__.co_filename)
_this_srcfile = __file__


class CustomIC(IceCreamDebugger):
    def __init__(self, logger):
        super().__init__()
        self.logger = logger

    def __call__(self, *args, **kwargs):
        if self.logger.level == logging.DEBUG:
            return super().__call__(*args, **kwargs)


class XposeLogger(logging.Logger):
    def __init__(self, name, level=logging.NOTSET):
        super().__init__(name, level)

    def findCaller(self, stack_info=None, stacklevel=None):
        f = currentframe()
        rv = "(unknown file)", 0, "(unknown function)"
        while f and hasattr(f, "f_code"):
            co, filename = f.f_code, os.path.normcase(f.f_code.co_filename)
            if co.co_name != 'test_logging' and filename in [_logging_srcfile, _this_srcfile]:
                f = f.f_back
                continue
            rv = (co.co_filename, f.f_lineno, co.co_name)
            break
        return rv

    def makeRecord(self, *args, **kwargs):
        record = super().makeRecord(*args, **kwargs)
        filename, lineno, funcName = self.findCaller()
        short_fn = os.path.basename(filename) if len(
            filename.split(os.sep)
            ) <= 2 else f"{filename.split(os.sep)[-3]}/{os.path.basename(filename)}"
        record.filename, record.lineno, record.funcName = short_fn, lineno, funcName
        return record


logging.setLoggerClass(XposeLogger)


class KafkaLoggingHandler(logging.Handler):
    def __init__(self, kafka_producer, topic_map):
        super().__init__()
        self.kafka_producer, self.topic_map = kafka_producer, topic_map

    def emit(self, record):
        topic = self.topic_map.get(record.levelname, 'debug_topic')
        log_dict = {k: getattr(record, k) for k in
                    ['message', 'levelname', 'name', 'filename', 'lineno', 'funcName', 'created']}
        self.kafka_producer.produce(topic, json.dumps(log_dict))


def get_logger(appConfig: ConfigModel):
    logger = logging.getLogger("xpose_logger")
    logger.setLevel(appConfig.log_to_console_loglevel)

    if appConfig.log_to_console_enabled:
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        ch.setFormatter(
            logging.Formatter(
                f'%(asctime)-25s | %(levelname)-8s | %(name)-15s | %(filename)-30s | %(funcName)-30s \n{"-" * 26}» %('
                f'levelname)-8s | %(message)s\n'
                )
            )
        logger.addHandler(ch)

    if appConfig.log_to_kafka_enabled:
        producer = Producer({'bootstrap.servers': appConfig.log_to_kafka_server_string})
        topic_map = {lvl: f"{lvl.lower()}_topic" for lvl in ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']}
        kh = KafkaLoggingHandler(producer, topic_map)
        kh.setLevel(logging.DEBUG)
        kh.setFormatter(
            logging.Formatter(
                '%(asctime)s - %(name)s - %(levelname)s - [%(filename)s::%(funcName)s] - %(message)s'
                )
            )
        logger.addHandler(kh)
        logger.debug(f"Logger initialized: {logger.name}")

    global ic
    ic = CustomIC(logger)
    return logger



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

class EventDispatcher:
    def __init__(self):
        self._events = {}

    def on_event(self, event_name):
        def decorator(func):
            self._events.setdefault(event_name, []).append(func)
            return func

        return decorator

    def dispatch(self, event_name, *args, **kwargs):
        for func in self._events.get(event_name, []):
            func(*args, **kwargs)


# Usage

dispatcher = EventDispatcher()


@dispatcher.on_event("startup")
def on_startup():
    print("Startup event triggered")


@dispatcher.on_event("shutdown")
def on_shutdown():
    print("Shutdown event triggered")


# Triggering events
dispatcher.dispatch("startup")
dispatcher.dispatch("shutdown")



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

import asyncio
import logging
import traceback
from typing import List

import uvicorn
from fastapi import APIRouter, FastAPI
from fastapi.middleware.cors import CORSMiddleware
from starlette.responses import JSONResponse
from uvicorn.config import LOGGING_CONFIG

from xposer.api.base.base_service import BaseService
from xposer.api.base.http.dto.response_wrapper_dto import ResponseWrapperDTO
from xposer.api.base.http.routes.debug_routes import XPDebugRouter
from xposer.core.context import Context

LOGGING_CONFIG["loggers"]["uvicorn"]["handlers"] = ["default"]
LOGGING_CONFIG["loggers"]["uvicorn.access"]["handlers"] = ["default"]
LOGGING_CONFIG["loggers"]["uvicorn.error"]["handlers"] = ["default"]
LOGGING_CONFIG["loggers"]["uvicorn.error"]["level"] = "ERROR"
LOGGING_CONFIG["loggers"]["xpose_logger"] = {
    "handlers": ["default"],
    "level": "DEBUG",
    }


class BaseFastApiService(BaseService):
    def __init__(self, ctx: Context):
        super().__init__(ctx)
        self.fastApi: FastAPI = None

    async def startService(
            self,
            host,
            port,
            routes: List[APIRouter],
            api_prefix: str = "/api",
            callback=None
            ):
        self.fastApi = FastAPI(tail_slash=True)

        self.fastApi.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
            )

        @self.fastApi.exception_handler(Exception)
        async def http_exception_handler(request, e):
            backtrace_info = traceback.format_exc()  # Gets the traceback information
            self.ctx.logger.error(
                f"FastAPI Internal error occurred @self.fastApi.exception_handler: {e}\n{backtrace_info}"
                )  # Logging the error with traceback
            detail = ResponseWrapperDTO(result="error", exception=str(e)).model_dump()
            return JSONResponse(status_code=500, content=detail)

        for route in routes:
            self.fastApi.include_router(route, prefix=api_prefix)

        if self.ctx.config.debug_enabled_for_built_in_http:
            debug_routes = [XPDebugRouter().getRoute(self.ctx)]
            for route in debug_routes:
                self.fastApi.include_router(route, prefix=api_prefix)

        @self.fastApi.on_event("startup")
        async def on_startup():
            if callback:
                self.ctx.logger.debug("FastAPI initialization successful")
                callback(None)

        xpose_logger = self.ctx.logger

        for uvicorn_logger_name in ["uvicorn", "uvicorn.error"]:
            uvicorn_specific_logger = logging.getLogger(uvicorn_logger_name)

            # Set log level for uvicorn.error specifically
            if uvicorn_logger_name == "uvicorn.error":
                uvicorn_specific_logger.setLevel(logging.INFO)

            uvicorn_specific_logger.handlers = xpose_logger.handlers
            uvicorn_specific_logger.propagate = False

        config = uvicorn.Config(
            app=self.fastApi,
            host=host,
            port=port,
            forwarded_allow_ips="*",
            log_level="debug",
            log_config=None
            )
        server = uvicorn.Server(config)
        server_task = asyncio.create_task(server.serve())
        server_task.set_name("Fastapi::Service")
        return server



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

import asyncio
import inspect
import json
from typing import Callable

from xposer.api.base.base_kafka_aiolib import AIOConsumer, AIOProducer
from xposer.api.base.base_service import BaseService
from xposer.core.context import Context


class BaseKafkaService(BaseService):

    def __init__(self, ctx: Context):
        super().__init__(ctx)
        _cancelled: bool = False

    async def start_service(
            self,
            server_string: str,
            group_id: str,
            inbound_topic: str,
            outbound_topic: str,
            exception_topic: str,
            handler_func: Callable = None,
            produce_on_result: bool = False
            ):

        self._cancelled = False

        if not inspect.iscoroutinefunction(handler_func):
            raise TypeError("handler_func must be an asynchronous coroutine function")

        consumer_config = {
            'bootstrap.servers': server_string,
            'group.id': group_id,
            'auto.offset.reset': 'latest',
            'fetch.message.max.bytes': '1048576',
            'fetch.wait.max.ms': '10',
            'heartbeat.interval.ms': '1000',
            'session.timeout.ms': '6000',
            'enable.auto.commit': False
            }
        producer_config = {
            'bootstrap.servers': server_string,
            }

        async def consumer_handler_func(data):
            """Warning: This is non thread-safe"""
            correlation_id = None
            if isinstance(data, dict):
                correlation_id = data.get('correlation_id', 'N/A')
            try:
                processed_data = await handler_func(data)
                response = {
                    'result': processed_data,
                    'correlation_id': correlation_id
                    }
                if produce_on_result:
                    try:
                        await self._producer.produce(outbound_topic, json.dumps(response))
                        self.ctx.logger.debug("Message produced successfully!")
                    except Exception as e:
                        self.ctx.logger.exception(f"Error producing message: {e}")
            except Exception as e:
                if exception_topic is not None:
                    exception_data = {
                        'exception': str(e),
                        'correlation_id': correlation_id
                        }
                    await self._producer.produce(exception_topic, json.dumps(exception_data))

        # Initiate AIOProducer and AIOConsumer
        self._producer = AIOProducer(producer_config)
        self._consumer = AIOConsumer(
            consumer_config,
            consumer_handler_func,
            inbound_topics=[inbound_topic],
            exception_queue=self.ctx.exception_queue
            )
        self.ctx.logger.info(f"Service initialized successfully. Listening on topic: {inbound_topic}.")

    async def stop_router(self):
        self.ctx.logger.info(f"Shutting down service")
        self._producer.close()
        self._consumer.close()

        # Stop all running asyncio tasks
        tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
        for task in tasks:
            task.cancel()



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

from pydantic import ConfigDict, Field
from pydantic_settings import BaseSettings


class BaseKafkaServiceConfigModel(BaseSettings):
    router_kafka_inbound_topic: str = Field(
        default="router_inbound",
        description="Default topicname the router should listen to"
        )
    router_kafka_outbound_topic: str = Field(
        default="router_outbound",
        description="Default topicname the router should publish results to"
        )
    router_kafka_exception_topic: str = Field(
        default="router_exception",
        description="Default topicname the router should throw exceptions"
        )
    router_kafka_server_string: str = Field(default="localhost:9092")
    router_kafka_group_id: str = Field(default="routers")

    model_config = ConfigDict(extra='allow')



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

from xposer.api.base.base_kafka_service import BaseKafkaService


class SampleAppKafkaService(BaseKafkaService):
    """Some example so developer may override"""
    ...



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

from fastapi import APIRouter, HTTPException
from pydantic import BaseModel, ValidationError

from xposer.api.base.http.decorators.wrap_response_decorator import ResponseWrapperDecorator
from xposer.core.context import Context


class SampleModel(BaseModel):
    name: str
    age: int


class SampleAppHTTPService:
    @staticmethod
    def getRoute(ctx: Context):
        router = APIRouter()

        @router.get("/test/")
        @ResponseWrapperDecorator(ctx)
        async def sample_route():
            return {
                "router": "custom",
                "xpcore": ctx.config.model_dump(),
                "xpcontroller": ctx.xpcontroller.config.model_dump()
                }

        @router.get("/test/exception")
        @ResponseWrapperDecorator(ctx)
        async def sample_route_with_exception():
            raise Exception("Test Exception")

        @router.get("/test/validation_error")
        @ResponseWrapperDecorator(ctx)
        async def sample_route_with_validation_error():
            data = {"name": "Alice", "age": "not_an_int"}  # Invalid data
            try:
                # This will raise a ValidationError
                valid_data = SampleModel.parse_obj(data)
            except ValidationError as e:
                raise HTTPException(status_code=400, detail=str(e))

            return valid_data

        return router



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

from abc import ABC
from typing import List

from pydantic import BaseModel

from xposer.api.base.base_service import BaseService
from xposer.core.abstract_xpcontroller import AbstractXPController
from xposer.core.context import Context


class XPControllerBaseClass(AbstractXPController, ABC):
    name: str = "XPControllerBaseClass"
    config_prefix: str = "xpcontroller_"  # Class level

    def __init__(self, ctx: Context):
        super().__init__(ctx)
        self.config = self.mergeConfigurationFromPrefix()
        # TODO self.socket_router: Any = None
        self.xpcontroller_conf_class: BaseModel
        self.services: List[BaseService] = []

    def mergeConfigurationFromPrefix(self) -> BaseModel:
        return BaseModel.model_construct()

    async def startXPController(self):
        raise NotImplementedError

    async def tearDownXPController(self):
        raise NotImplementedError

    async def asyncInit(self):
        ...



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

import __main__
import os
import re
# Existing imports
from platform import python_version

import pkg_resources
import psutil
from fastapi import APIRouter, Body

from xposer.core.context import Context


def get_root_package_version():
    root_package_name = __main__.__package__
    if root_package_name:
        try:
            return pkg_resources.get_distribution(root_package_name).version
        except pkg_resources.DistributionNotFound:
            pass

    # Fallback: Look for setup.py in the main script directory
    main_script_dir = os.path.dirname(os.path.abspath(__main__.__file__))
    setup_py_path = os.path.join(main_script_dir, 'setup.py')
    if os.path.exists(setup_py_path):
        with open(setup_py_path, 'r') as f:
            setup_content = f.read()
            version_match = re.search(r'version=["\']([^"\']+)["\']', setup_content)
            if version_match:
                return version_match.group(1)

    return "Not Applicable"


class XPDebugRouter:
    @staticmethod
    def getRoute(ctx: Context):
        router = APIRouter()

        @router.post("/echo/")
        async def echo(params: dict = Body(...)):
            return {"echo": params}

        @router.get("/debug")
        async def read_debug_info():
            python_version_info = python_version()
            current_package_version = pkg_resources.get_distribution("xposer").version
            all_packages = {d.project_name: d.version for d in pkg_resources.working_set}
            memory_info = {
                "virtual_memory": psutil.virtual_memory()._asdict(),
                "swap_memory": psutil.swap_memory()._asdict(),
                }

            root_package_version = get_root_package_version()  # Get root package version

            return {
                "python_version": python_version_info,
                "root_package_version": root_package_version,  # Include root package version
                "xposer": current_package_version,
                "all_packages": all_packages,
                "memory_info": memory_info,
                "configuration": ctx.config.model_dump()
                }

        return router



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

import asyncio
import json
from concurrent.futures import ThreadPoolExecutor
from queue import Queue

from confluent_kafka import Consumer, KafkaException, Producer


class AIOProducer:
    def __init__(self, configs, loop=None):
        self._loop = loop or asyncio.get_event_loop()
        self._producer = Producer(configs)
        self._cancelled = False
        self._poll_task = self._loop.create_task(self._poll_loop())
        self._poll_task.set_name("AIOProducer::Poll")

    async def _poll_loop(self):
        while not self._cancelled:
            self._producer.poll(0.1)
            await asyncio.sleep(0.1)

    def close(self):
        self._cancelled = True
        self._producer.flush()
        if self._poll_task:  # Ensure the task is available before trying to cancel
            self._poll_task.cancel()
            self._producer.flush()

    async def produce(self, topic, value):
        result = self._loop.create_future()

        def ack(err, msg):
            if err:
                result.set_exception(KafkaException(err))
            else:
                result.set_result(msg)

        self._producer.produce(topic, value, on_delivery=ack)

        return await result


class AIOConsumer:
    def __init__(
            self,
            configs,
            handler_func,
            inbound_topics: list,
            exception_queue: Queue,
            loop=None
            ):
        self._loop = loop or asyncio.get_event_loop()
        self._consumer = Consumer(configs)
        self._exception_queue = exception_queue
        self._consumer.subscribe(inbound_topics)
        self._handler_func = handler_func
        self._cancelled = False
        self._executor = ThreadPoolExecutor(max_workers=1)
        self._consume_task = self._loop.create_task(self._consume_loop())
        self._consume_task.set_name("AIOConsumer::Consume")

    async def _handle_and_commit(self, msg):
        raw_value = msg.value().decode('utf-8') if msg.value() else None
        try:
            data = json.loads(raw_value)
        except json.JSONDecodeError:
            data = raw_value  # Continue with the raw string if it's not JSON-decodable

        if not data:
            raise Exception("Missing value from the message", msg)

        await self._handler_func(data)
        self._consumer.commit(message=msg)

    async def _consume_loop(self):
        while not self._cancelled:
            try:
                msg = await self._loop.run_in_executor(self._executor, self._consumer.poll, 0.1)
                if msg:
                    await self._handle_and_commit(msg)
            except Exception as e:
                if self._exception_queue:
                    self._exception_queue.put_nowait(e)

    def close(self):
        self._cancelled = True
        self._executor.shutdown()
        if self._consume_task:
            self._consume_task.cancel()



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

# cli_interface.py

import argparse


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("operation", choices=['add', 'subtract'])
    parser.add_argument("x", type=float)
    parser.add_argument("y", type=float)

    args = parser.parse_args()

    calculator = Calculator()

    if args.operation == 'add':
        print(calculator.add(args.x, args.y))
    elif args.operation == 'subtract':
        print(calculator.subtract(args.x, args.y))


if __name__ == "__main__":
    main()



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

from typing import Dict, Optional

from pydantic import BaseModel


class RequestMetaDTO(BaseModel):
    skip: Optional[int] = None
    limit: Optional[int] = None
    fields: Optional[Dict[str, int]] = None



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

import asyncio
import queue
import signal
import sys
import threading
import traceback
from typing import Any

from xposer.core.completed_exception import CompletedException
from xposer.core.configure import Configurator
from xposer.core.context import Context
from xposer.core.logger import get_logger
from xposer.core.xpcontroller_factory import XPControllerFactory
from xposer.core.xpose_task import XPTask


class Boot:
    def __init__(self) -> None:
        self.ctx = None  # type: Context
        self.shutdown_event = asyncio.Event()
        self.shutdown_in_progress = False

    def _sync_shutdown_handler(self, *_: Any) -> None:
        self.ctx.logger.warning("\n*** CLI/Sync Shutdown signal received ***")
        if not self.shutdown_in_progress:
            try:
                self.shutdown_in_progress = True
                asyncio.get_running_loop()
                asyncio.create_task(self.shutdown())
            except RuntimeError:
                sys.exit()

    def thread_exception_handler(self, args):
        try:
            thread_name = args.thread.name if args.thread else "Unknown"
            self.ctx.logger.error(f"Exception in thread: {thread_name}")
            tb = args.exc_value.__traceback__
            traceback.print_exception(args.exc_type, args.exc_value, tb)
            if threading.current_thread().name == "MainThread":
                self._sync_shutdown_handler()
            else:
                if self.ctx.exception_queue:
                    self.ctx.exception_queue.put_nowait(args)
                else:
                    sys.exit(0)
        except Exception as e:
            print(f"Failed to handle exception in thread: {e}")

    async def monitor_exceptions(self) -> None:
        while not self.shutdown_event.is_set():
            try:
                exception = self.ctx.exception_queue.get_nowait()
                if isinstance(exception, CompletedException):
                    self.ctx.logger.info(f"Controller completed: {exception.args[0]}")
                elif hasattr(exception, 'exc_type') and exception.exc_type is CompletedException:
                    self.ctx.logger.info(f"Controller completed: {exception.args[0]}")
                else:
                    self.ctx.logger.error(f"Exception: {exception}")
                    try:
                        tb = exception.exc_value.__traceback__
                        traceback.print_exception(exception.exc_type, exception.exc_value, tb)
                    except Exception:
                        self.ctx.logger.warning("Unable to print monitor_exceptions exception due to ")
                await self.shutdown()
            except (queue.Empty, asyncio.CancelledError):
                await asyncio.sleep(0.1)

    async def shutdown(self) -> None:
        if self.shutdown_in_progress:
            return
        self.ctx.logger.warning("*** Internal Shutdown signal received ***")
        self.shutdown_in_progress = True
        self.ctx.logger.info("Shutting down application initiated")
        self.ctx.message_queue.put({'target': None, 'message': 'shutdown'})
        await asyncio.sleep(1)

        tasks = [asyncio.create_task(xptask.shutdown()) for xptask in self.ctx.xptask_list]
        self.ctx.logger.debug(f"Shutting down XPTasks")
        await asyncio.gather(*tasks)
        self.ctx.logger.debug(f"Shutting down Main loop")
        await XPTask.cancel_tasks_for_loop(self.ctx, asyncio.get_event_loop())
        self.ctx.logger.debug(f"Main shutdown sequence completed")
        sys.exit()

    def handle_loop_exceptions(self, loop: asyncio.AbstractEventLoop, context: dict) -> None:
        task = context.get('future')
        if task and (exc := task.exception()):
            if isinstance(exc, asyncio.CancelledError) and self.shutdown_in_progress:
                return  # Skip logging and shutdown for CancelledError when shutdown is in progress

            self.ctx.logger.error(f"Loop exception: {exc}")

            if not self.shutdown_in_progress:
                loop.run_until_complete(self.shutdown())

    async def boot(self) -> Context:
        config = Configurator.buildConfig()
        logger = get_logger(config)
        context = Context(logger, config, {})
        self.ctx = context
        loop = asyncio.get_event_loop()
        loop.id = "main_loop"
        xpcontroller = XPControllerFactory.make(context)
        await xpcontroller.asyncInit()
        asyncio.get_event_loop().set_exception_handler(self.handle_loop_exceptions)
        threading.excepthook = self.thread_exception_handler
        xptask_initialization_future = asyncio.Future()
        xptask = XPTask(self.ctx)

        def initialization_callback():
            loop.call_soon_threadsafe(xptask_initialization_future.set_result, None)

        logger.info(f"Boot sequence completed. Starting XPController {xpcontroller.name}")
        xptask.startup(
            to_be_threadified_func=xpcontroller.startXPController,
            initialization_callback=initialization_callback,
            teardown_func=xpcontroller.tearDownXPController,
            main_event_loop=loop,
            exception_callback=None,
            custom_logger=logger,
            task_slug='boot_xpcontroller_startXPController',
            re_raise_exception=True
            )
        await xptask_initialization_future
        self.ctx.xptask_list.append(xptask)
        context.xpcontroller = xpcontroller

        for s in (signal.SIGTERM, signal.SIGINT):
            signal.signal(s, self._sync_shutdown_handler)

        logger.info(f"XPController {xpcontroller.name} sync process complete. Async tasks might be in progress.")

        try:
            monitor_task = asyncio.create_task(self.monitor_exceptions())
            shutdown_task = asyncio.create_task(self.shutdown_event.wait())
            await asyncio.gather(monitor_task, shutdown_task)
        except asyncio.CancelledError as ce:
            if self.shutdown_in_progress:
                self.ctx.logger.debug("CancelledError caught during shutdown. This is expected.")
            else:
                self.ctx.logger.error(f"Exception: {ce}")
                await self.shutdown()
                raise ce
        except Exception as e:
            self.ctx.logger.error(f"Exception: {e}")
            if not self.shutdown_in_progress:
                await self.shutdown()
            raise e

        return context



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

import queue
from logging import Logger
from typing import Any, Dict, List

from xposer.core.configuration_model import ConfigModel


class Context:
    _instance = None
    logger: Logger = None
    config: ConfigModel = None
    message_queue: queue.Queue[Dict[str, Any]] = queue.Queue()
    exception_queue: queue.Queue[Dict[str, Any]] = queue.Queue()
    xptask_list: List[Any] = []
    state = None
    xpcontroller: None

    def __init__(self, logger, config, state):
        self.logger = logger
        self.config = config
        self.state = state



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

import asyncio

from pydantic import Field, field_validator

from xposer.api.base.base_fastapi_service import BaseFastApiService
from xposer.api.base.base_fastapi_service_config_model import BaseFastApiRouterConfigModel
from xposer.api.base.xpcontroller_base_class import XPControllerBaseClass
from xposer.core.configure import Configurator
from xposer.core.context import Context
from xposer.sample_app.http_post_uvicorn_fastapi.routers.sample_app_http_service import SampleAppHTTPService


class SampleAppHttpControllerConfigModel(BaseFastApiRouterConfigModel):
    foo_param: str = Field(default='foo')
    uvicorn_host: str = Field(default='localhost')
    uvicorn_port: int = Field(default=8000)

    @field_validator("uvicorn_port")
    def convert_to_int(cls, value):
        return int(value)


class SampleAppHttpXPController(XPControllerBaseClass):
    config_prefix: str = "xpcontroller_"  # Class level

    def __init__(self, ctx: Context):
        super().__init__(ctx)
        self.api_prefix: str = "/api"
        self.uvicorn_server = None
        self.http_router: BaseFastApiService = None
        self.config: SampleAppHttpControllerConfigModel = self.config  # Type hint

    def mergeConfigurationFromPrefix(self) -> SampleAppHttpControllerConfigModel:
        return Configurator.mergeAttributesWithPrefix(
            SampleAppHttpControllerConfigModel,
            self.ctx.config,
            self.config_prefix,
            validate=True,
            strict=True
            )

    async def start_fastapi_service(self, callback):
        try:
            self.uvicorn_server = await self.http_router.startService(
                self.config.uvicorn_host,
                self.config.uvicorn_port,
                [SampleAppHTTPService.getRoute(self.ctx)],
                api_prefix=self.api_prefix,
                callback=callback,
                )
        except Exception as e:
            # Log the exception for debugging
            self.ctx.logger.exception(f"Error starting FastApi service: {e}")
            # Notify the future object about the failure
            raise

    async def tearDownXPController(self):
        self.ctx.logger.info("tearDownXPController called")
        if self.uvicorn_server:
            self.uvicorn_server.should_exit = True
        await asyncio.sleep(1)

    def handle_timeout_exception(self, task):
        try:
            task.result()
        except asyncio.TimeoutError:
            raise ValueError("The FastAPI service did not start within 10 seconds!")

    async def startXPController(self):
        # raise CompletedException(self.__class__.__name__)
        self.http_router = BaseFastApiService(self.ctx)
        future = asyncio.Future()
        fastapi_service_task = asyncio.create_task(self.start_fastapi_service(callback=future.set_result))
        fastapi_service_task.set_name("SampleAppHttpController:StartFastApiServiceTask")
        timeout_task = asyncio.create_task(asyncio.wait_for(future, timeout=10))
        timeout_task.set_name("SampleAppHttpXPController::FastApiServiceTimeoutTask")
        timeout_task.add_done_callback(self.handle_timeout_exception)
        await future
        self.ctx.logger.debug("Started")



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

from typing import Any

from xposer.core.context import Context


class BaseService:
    app: Any = None

    def __init__(self, ctx: Context):
        self.ctx = ctx

    async def startService(self, *args, **kwargs):
        raise NotImplementedError("Subclasses must implement this method to define their service")

    async def stopService(self, *args, **kwargs):
        raise NotImplementedError("Subclasses must implement this method to define their service")



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

import asyncio
import inspect
import logging
import sys
import threading
from asyncio import Event
from typing import Any, Callable, Optional, TypeVar, Union

import shortuuid

from xposer.core.context import Context
from xposer.core.coro_exception import CoroException

T = TypeVar('T', bound='MyClass')


class XPTask:

    @staticmethod
    async def cancel_tasks_for_loop(ctx, loop, timeout=1):
        current_loop = asyncio.get_running_loop()
        pending = asyncio.all_tasks(loop=loop)  # Get all pending tasks
        if len(pending) == 0:
            await loop.shutdown_asyncgens()
            return
        try:
            # Cancel all pending tasks
            for task in pending:
                task.cancel()

            # If the loop is the current loop, await the tasks in the current loop
            if loop == current_loop:
                await asyncio.wait_for(asyncio.gather(*pending, return_exceptions=True), timeout)

            # If the loop is not the current loop, run in a threadsafe manner
            else:
                future = asyncio.run_coroutine_threadsafe(
                    asyncio.wait_for(asyncio.gather(*pending, return_exceptions=True), timeout), loop
                    )
                return future.result(timeout=timeout)
        except asyncio.TimeoutError:
            ctx.logger.warn(f"Tasks did not cancel within {timeout} seconds.")
        except asyncio.CancelledError:
            ctx.logger.warn("Task was cancelled, this is expected behavior.")
        except Exception as e:
            print(e)  # Handle other exceptions
        finally:
            await loop.shutdown_asyncgens()

        # Recheck for remaining tasks
        remaining = asyncio.all_tasks(loop=loop)

        if remaining:
            ctx.logger.warn(f"{len(remaining)} tasks have not terminated yet.")
            for task in remaining:
                ctx.logger.warn(f"Unterminated task name: {task.get_name()}")
        else:
            ctx.logger.warn("All tasks have been canceled.")

        # Safely close the loop
        if loop.is_running():
            # ctx.logger.warn("Loop is still running. Can't close it now.")
            ...
        else:
            loop.close()

    def __init__(self, ctx: Context) -> None:
        self.re_raise_exception: Optional[bool] = None
        self.wrapped_threaded_func_task: Optional[asyncio.Task] = None
        self.wrapped_threaded_func_task_slug: Optional[str] = None
        self.wrapped_threaded_func_task_loop: Optional[asyncio.AbstractEventLoop] = None
        self.custom_logger: Optional[logging.Logger] = None
        self.shutdown_event: Event = None
        self.id = shortuuid.ShortUUID().random(length=7)
        task: Optional[asyncio.Task]
        ctx: Context
        task_loop: Optional[asyncio.AbstractEventLoop]
        is_shut_down: bool = False
        self.ctx = ctx

    @property
    def logger(self) -> logging.Logger:
        return self.custom_logger or logging.getLogger(self.wrapped_threaded_func_task_slug or "XPTask")

    def handle_task_loop_exception(self, loop: asyncio.AbstractEventLoop, context: dict) -> None:
        self.logger.error(f"[XPTask] Caught exception in {self.wrapped_threaded_func_task_slug}: {context['message']}")
        exception = context.get('exception')
        if self.ctx.exception_queue:
            self.ctx.exception_queue.put_nowait(exception)

    def run_loop_in_thread(self, to_be_threadified_func: Callable) -> None:
        async def coroutine_wrapper():
            try:
                if inspect.iscoroutinefunction(to_be_threadified_func):
                    await to_be_threadified_func()
                else:
                    to_be_threadified_func()
            except Exception as e:
                if self.ctx.exception_queue:
                    exc_type, exc_value, tb = sys.exc_info()
                    exception = CoroException(exc_type, exc_value, tb)
                    self.ctx.exception_queue.put_nowait(exception)
            finally:
                self.initialization_callback()

        # Create a new event loop
        self.wrapped_threaded_func_task_loop = asyncio.new_event_loop()
        self.wrapped_threaded_func_task_loop.id = self.id

        # Set the event loop for this context
        asyncio.set_event_loop(self.wrapped_threaded_func_task_loop)

        # Set exception handler for indefinitely running create_task or gather commands
        self.wrapped_threaded_func_task_loop.set_exception_handler(self.handle_task_loop_exception)

        # Create a new task to run the provided function

        self.wrapped_threaded_func_task = self.wrapped_threaded_func_task_loop.create_task(coroutine_wrapper())
        self.wrapped_threaded_func_task.set_name("XPTASK:WrappedThreadedFunctionCoroutine")

        # Run the event loop indefinitely
        # self.wrapped_threaded_func_task_loop.run_forever()
        self.wrapped_threaded_func_task_loop.run_forever()

    def startup(
            self,
            to_be_threadified_func: Callable,
            initialization_callback: Callable,
            teardown_func: Callable,
            main_event_loop: asyncio.AbstractEventLoop,
            exception_callback: Callable[[Union[Any, None], Exception], None],
            custom_logger: Optional[logging.Logger] = None,
            task_slug: str = '',
            re_raise_exception: bool = True
            ) -> T:
        self.main_event_loop = main_event_loop
        self.on_exception_callback = exception_callback
        self.initialization_callback = initialization_callback
        self.custom_logger = custom_logger
        self.teardown_func = teardown_func
        self.re_raise_exception = re_raise_exception
        self.wrapped_threaded_func_task_slug = task_slug
        self.logger.debug("Creating task in a new thread")
        self.thread = threading.Thread(target=self.run_loop_in_thread, args=(to_be_threadified_func,))

        self.thread.daemon = True
        self.thread.start()

    async def shutdown(self):
        # Todo implement graceful options for XPTasks that are not actually daemons
        self.ctx.logger.debug(f"Shutting down XPtask:{self.id}")
        if inspect.iscoroutinefunction(self.teardown_func):
            await self.teardown_func()
        else:
            self.teardown_func()
        await XPTask.cancel_tasks_for_loop(self.ctx, self.wrapped_threaded_func_task_loop)
        """
        if self.thread.is_alive():
            self.ctx.logger.warning(f"Terminating thread for XPtask:{self.id}")
        self.thread.join()
        pass
        """
        ...

    def __del__(self) -> None:
        """Close the event loop if it is not already closed."""
        self.shutdown()



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

from typing import Any, Optional

from pydantic import BaseModel


class ResponseStatsDTO(BaseModel):
    total_count: int = None
    has_mode: Optional[bool] = None


class ResponseDTO(BaseModel):
    items: Any = None
    stats: Optional[ResponseStatsDTO] = None



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

import importlib
import json
from typing import cast

from xposer.core.abstract_xpcontroller import AbstractXPController
from xposer.core.context import Context


class XPControllerFactory:

    @staticmethod
    def make(ctx: Context) -> AbstractXPController:
        # Find and map slug to classnames
        xpcontroller_module_name = ctx.config.model_dump().get("xpcontroller_module_name", None)
        if xpcontroller_module_name is None:
            raise EnvironmentError(f"Unable to resolve xpcontroller module name from config")
        xpcontroller_class_name = ctx.config.model_dump().get("xpcontroller_class_name", None)
        if xpcontroller_class_name is None:
            raise EnvironmentError(f"Unable to resolve xpcontroller class name from config")

        # Instantiate the resolved class
        try:
            module = importlib.import_module(xpcontroller_module_name)
            class_ = getattr(module, xpcontroller_class_name)
            ctx.logger.debug(f"Trying to load module:{xpcontroller_module_name} class:{xpcontroller_class_name}")
            instance: AbstractXPController = cast(AbstractXPController, class_(ctx))
        except ImportError as e:
            raise ImportError(f"Module {xpcontroller_module_name} not found. Original error: {json.dumps(str(e))}")
        except AttributeError as e:
            raise AttributeError(f"Class {xpcontroller_class_name} not found in the module. Original error: {str(e)}")
        return instance



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

import logging

from pydantic import ConfigDict, Field, field_validator
from pydantic_settings import BaseSettings


class ConfigModel(BaseSettings):
    debug_enabled_for_built_in_http: bool = Field(
        default=True,
        error_messages={
            }
        )
    log_to_console_enabled: bool = Field(
        default=True,
        error_messages={
            }
        )
    log_to_console_loglevel: int | str = Field(
        logging.DEBUG,
        error_messages={
            }
        )
    log_to_kafka_enabled: bool = Field(
        default=True,
        error_messages={
            }
        )
    log_to_kafka_server_string: str = Field(
        'localhost:9092',
        error_messages={
            }
        )
    log_to_kafka_server_log_topic: str = Field(
        'log',
        error_messages={
            }
        )
    xpcontroller_module_name: str | None = Field(
        ...,
        error_messages={
            }
        )
    xpcontroller_class_name: str | None = Field(
        ...,
        error_messages={
            }
        )

    model_config = ConfigDict(extra='allow', arbitrary_types_allowed=True)

    @field_validator('log_to_console_loglevel')
    def map_log_level(cls, value):
        if isinstance(value, str):
            level = getattr(logging, value.upper(), None)
            if isinstance(level, int):
                return level
            else:
                raise ValueError("Invalid log level name.")
        elif isinstance(value, int):
            return value
        else:
            raise TypeError("Invalid log level type.")



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

import argparse
import os
import warnings
from typing import Any, Type, TypeVar
from typing import Dict, Optional, Union

import yaml
from pydantic import BaseModel
from pydantic_settings import BaseSettings

from xposer.core.configuration_model import ConfigModel

T = TypeVar('T')


class Configurator:

    @staticmethod
    def convert_to_bool(value: Union[bool, str, int]) -> bool:
        """Convert various representations of boolean to actual boolean."""
        value_map = {
            'true': True, 'True': True, '1': True, 1: True,
            'false': False, 'False': False, '0': False, 0: False
            }
        return value_map.get(value, value)

    def normalize_bool_fields(config_model: Union[BaseSettings, BaseModel]):
        """Filter bool type str warning (we are fixing it below so it's expected)"""
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            for field_name, field_value in config_model.model_dump().items():
                field_type = config_model.__annotations__.get(field_name, None)

                # Handle nested Pydantic models
                if isinstance(field_value, (BaseSettings, BaseModel)):
                    Configurator.normalize_bool_fields(field_value)

                # Handle lists, traverse into dicts or Pydantic models
                elif isinstance(field_value, list):
                    new_list = []
                    for item in field_value:
                        if isinstance(item, dict):
                            new_dict = {
                                k: Configurator.convert_to_bool(v) if isinstance(v, str) and v.lower() in ['true',
                                                                                                           'false']
                                else v
                                for k, v in item.items()}
                            new_list.append(new_dict)
                        elif isinstance(item, (BaseSettings, BaseModel)):
                            Configurator.normalize_bool_fields(item)
                            new_list.append(item)
                        else:
                            new_list.append(item)
                    setattr(config_model, field_name, new_list)

                # Convert bool fields
                elif field_type is bool:
                    setattr(config_model, field_name, Configurator.convert_to_bool(field_value))
            return config_model

    @staticmethod
    def shallow_dict_from_object_with_prefix_removal(
            source: Union[Any, Dict[str, Any]],
            prefix: str = '',
            strict: bool = False
            ) -> Dict[str, Any]:

        if hasattr(source, 'dict'):
            source_data = source.dict()
        elif isinstance(source, dict):
            source_data = source
        else:
            raise ValueError("source must be an instance with a dict() method or a dictionary.")

        output = {}
        for key, value in source_data.items():
            if key.startswith(prefix):
                output[key[len(prefix):]] = value
            elif not strict:
                output[key] = value

        return output

    @staticmethod
    def prefilled_dict_from_class_and_object(
            target_cls: Union[Type[BaseModel], Type[BaseSettings], Dict[str, Any]],
            source_obj: Optional[
                Union[BaseModel, BaseSettings, Dict[str, Any]]] = None,
            strict: bool = False
            ) -> Dict[str, Any]:
        if issubclass(target_cls, (BaseModel, BaseSettings)):
            shallow_target_dict = {k: v.get_default() for k, v in
                                   target_cls.model_fields.items()}
        elif isinstance(target_cls, dict):
            shallow_target_dict = target_cls
        else:
            raise ValueError(
                "target_cls must be either a BaseModel-derived class, BaseSettings-derived class, or a dictionary."
                )

        if source_obj:
            if isinstance(source_obj, (BaseModel, BaseSettings)):
                source_data = source_obj.model_dump()
            elif isinstance(source_obj, dict):
                source_data = source_obj
            else:
                raise ValueError("source_obj must be either a BaseModel-derived instance or a dictionary.")

            for key, value in source_data.items():
                if strict:
                    if key in shallow_target_dict:
                        shallow_target_dict[key] = value
                else:
                    shallow_target_dict[key] = value

        return shallow_target_dict

    @staticmethod
    def prefilled_dict_from_object_and_object(
            target_obj: Union[BaseModel, Dict[str, Any]],
            source_obj:
            Union[BaseModel, BaseSettings, Dict[str, Any]] = None,
            strict: bool = False
            ) -> Dict[str, Any]:
        if isinstance(target_obj, (BaseModel, BaseSettings)):
            shallow_target_dict = {k: v.get_default() for k, v in
                                   target_obj.model_fields.items()}
            for k in shallow_target_dict:
                shallow_target_dict[k] = getattr(target_obj, k, None)
            for k, v in target_obj.model_extra.items():
                shallow_target_dict[k] = v

        elif isinstance(target_obj, dict):
            shallow_target_dict = target_obj
        else:
            raise ValueError(
                "target_cls must be either a BaseModel-derived class instance, BaseSettings-derived class instance, "
                "or a dictionary instance."
                )

        if isinstance(source_obj, (BaseModel, BaseSettings)):
            source_data = source_obj.model_dump()
        elif isinstance(source_obj, dict):
            source_data = source_obj
        else:
            raise ValueError("source_obj must be either a BaseModel-derived instance or a dictionary intance.")

        for key, value in source_data.items():
            if strict:
                if key in shallow_target_dict:
                    shallow_target_dict[key] = value
            else:
                shallow_target_dict[key] = value

        return shallow_target_dict

    @staticmethod
    def mergeAttributesWithPrefix(
            target: Type[T],
            source: Union[BaseModel, Dict[str, Any]],
            prefix: str = '',
            validate: bool = True,
            strict: bool = True
            ) -> T:

        source_dict = Configurator.shallow_dict_from_object_with_prefix_removal(source, prefix, strict=False)
        result_obj = None
        # Determine whether the target is a class or an object and get the prefilled dictionary
        if isinstance(target, type):
            merged_dict = Configurator.prefilled_dict_from_class_and_object(target, source_dict, strict)
            if issubclass(target, Dict):
                result_obj = merged_dict
            else:
                result_obj = target.model_construct(**merged_dict)

        else:
            merged_dict = Configurator.prefilled_dict_from_object_and_object(target, source_dict, strict)
            if isinstance(target, Dict):
                result_obj = type(target)(**merged_dict)
            else:
                result_obj = type(target).model_construct(**merged_dict)

        # Validate the object if required
        # Pydantic has a model_validate classmethod, we focus on those cases
        bool_normalized_obj = Configurator.normalize_bool_fields(result_obj)

        if validate:
            if isinstance(target, type):
                if issubclass(target, Dict):
                    # dont validate a dict type
                    pass
                else:
                    result_obj = target(**bool_normalized_obj.model_dump())
            else:
                if isinstance(target, Dict):
                    # dont validate a dict instance
                    pass
                elif isinstance(target, (BaseModel, BaseSettings)):
                    result_obj = type(target)(**bool_normalized_obj.model_dump())

        return result_obj

    @staticmethod
    def parseConfig(config_filename):

        if config_filename is None:
            raise EnvironmentError("  environment variable is not set.")

        if not os.path.exists(config_filename):
            raise FileNotFoundError(f"Configuration file '{config_filename}' not found.")

        with open(config_filename, 'r') as file:
            return yaml.safe_load(file)

    @staticmethod
    def getVarsFromCLIArgs(parser: argparse.ArgumentParser) -> Dict:
        args, otherargs = parser.parse_known_args()

        # Convert recognized arguments to dictionary
        args_dict = vars(args)

        # Process otherargs to extract key-value pairs
        otherargs_dict = {}
        for item in otherargs:
            if "=" in item:
                key, value = item.lstrip('-').split('=', 1)
                otherargs_dict[key] = value
            else:
                otherargs_dict[item.lstrip('-')] = True

        # Merge dictionaries
        return {**args_dict, **otherargs_dict}

    @staticmethod
    def buildConfig() -> ConfigModel:
        ROOT_PREFIX: str = 'xp_'
        # Determine configuration_filename source and value
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "--config",
            type=str,
            help="Config file path",
            required=False,
            default=None
            )
        args, otherargs = parser.parse_known_args()
        config_filename = args.config
        if os.environ.get("XPOSER_CONFIG", None) is not None:
            config_filename = os.environ.get("XPOSER_CONFIG")

        if config_filename is None:
            raise EnvironmentError(f"Missing environment variable for building context: {config_filename}")

        # remove --config from arguments list so it will not
        # Load configuration file
        loaded_and_parsed_config = Configurator.parseConfig(config_filename)

        # Merge initially loaded values from configuration file
        config_file_configuration: ConfigModel = Configurator.mergeAttributesWithPrefix(
            ConfigModel,
            loaded_and_parsed_config,
            ROOT_PREFIX,
            validate=False,
            strict=False
            )

        # Override configuration from variables from environment
        environment_overridden_configuration: ConfigModel = Configurator.mergeAttributesWithPrefix(
            config_file_configuration,
            {key.lower(): value for key, value
             in os.environ.items()},
            ROOT_PREFIX,
            validate=False,
            strict=False
            )

        cli_args_with_prefix = Configurator.getVarsFromCLIArgs(parser)

        # Merge all variable from cli args (override where applicable)
        cli_overridden_configuration: ConfigModel = Configurator.mergeAttributesWithPrefix(
            environment_overridden_configuration,
            cli_args_with_prefix,
            ROOT_PREFIX,
            validate=False,
            strict=False
            )

        # Normalize bool fields
        bool_normalized_configuration = Configurator.normalize_bool_fields(cli_overridden_configuration)

        # Now do the validation business
        validated_configuration = ConfigModel(**bool_normalized_configuration.model_dump(), strict=True)
        return validated_configuration



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

import traceback
from functools import wraps

from fastapi import HTTPException

from xposer.api.base.http.dto.response_wrapper_dto import ResponseWrapperDTO


def ResponseWrapperDecorator(ctx):
    def actual_decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            try:
                result = await func(*args, **kwargs)
                return ResponseWrapperDTO(result="success", data=result)
            except Exception as e:
                backtrace_info = traceback.format_exc()  # Gets the traceback information
                ctx.logger.error(
                    f"FastAPI Internal error occurred @ResponseWrapperDecorator: {e}\n{backtrace_info}"
                    )  # Logging the error with traceback
                detail = ResponseWrapperDTO(result="error", exception=str(e)).model_dump()
                raise HTTPException(status_code=500, detail=detail)

        return wrapper

    return actual_decorator



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

from abc import ABC, abstractmethod
from typing import Generic, Type, TypeVar

from xposer.core.context import Context

T = TypeVar('T')


class AbstractXPController(ABC, Generic[T]):
    xpcontroller_conf_class: Type[T]
    config_prefix: str

    def __init__(self, ctx: Context):
        self._ctx = ctx

    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @name.setter
    @abstractmethod
    def name(self, value: str) -> None:
        ...

    @property
    def ctx(self) -> Context:
        return self._ctx

    @ctx.setter
    def ctx(self, value: Context) -> None:
        self._ctx = value

    @abstractmethod
    async def startXPController(self) -> None:
        ...

    @abstractmethod
    async def tearDownXPController(self) -> None:
        ...

    @abstractmethod
    def mergeConfigurationFromPrefix(self) -> T:
        ...

    @abstractmethod
    async def asyncInit(self) -> None:
        ...



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

from typing import Any

from pydantic import BaseModel


class ResponseWrapperDTO(BaseModel):
    result: str
    exception: Any = None
    data: Any = None



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

from pydantic import ConfigDict, Field
from pydantic_settings import BaseSettings


class BaseFastApiRouterConfigModel(BaseSettings):
    model_config = ConfigDict(extra='allow')



from setuptools import find_packages, setup


def read_requirements():
    with open('requirements.txt', 'r') as f:
        requirements = f.read().splitlines()
    return requirements


setup(
    name='xposer',
    version='1.0.1',
    packages=find_packages(),
    install_requires=read_requirements(),
    author='Aron Barocsi',
    author_email='aron.barocsi@gmail.com',
    description='Xpose your functions as microservices over arbitrary channels using standardized logging and configuration',
    license='Private',
    keywords='microservice, function wrapper, function exposer',
    url='https://github.com/barocsi/xposer',
    )



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

from abc import ABC, abstractmethod


class AbstractHandler(ABC):
    @abstractmethod
    def handle(self, context, *args, **kwargs):
        """"""



#  Copyright (c) 2024. Aron Barocsi | All rights reserved.

class CoroException(Exception):
    def __init__(self, exc_type, exc_value, tb):
        self.exc_type = exc_type
        self.exc_value = exc_value
        self.tb = tb
